# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import logging
import re

import jsonpath_rw as jsonpath
import requests
from lxml import etree

from eodag.api.product import EOProduct
from eodag.api.product.metadata_mapping import (
    format_metadata,
    get_search_param,
    properties_from_json,
    properties_from_xml,
)
from eodag.plugins.search.base import Search
from eodag.utils import parse_qs, urlencode
from eodag.utils.exceptions import RequestError

logger = logging.getLogger("eodag.plugins.search.qssearch")


class QueryStringSearch(Search):
    """A plugin that helps implementing any kind of search protocol that relies on
    query strings (e.g: opensearch).

    The available configuration parameters for this kind of plugin are:

        - **result_type**: (optional) One of "json" or "xml", depending on the
          representation of the provider's search results. The default is "json"

        - **results_entry**: (mandatory) The name of the key in the provider search
          result that gives access to the result entries

        - **api_endpoint**: (mandatory) The endpoint of the provider's search interface

        - **literal_search_params**: (optional) A mapping of (search_param =>
          search_value) pairs giving search parameters to be passed as is in the search
          url query string. This is useful for example in situations where the user wants
          to pass-in a search query as it is done on the provider interface. In such a case,
          the user can put in his configuration file the query he needs to pass to the provider.

        - **pagination**: (mandatory) The configuration of how the pagination is done
          on the provider. It is a tree with the following nodes:

          - *next_page_url_tpl*: The template for pagination requests. This is a simple
            Python format string which will be resolved using the following keywords:
            ``url`` (the base url of the search endpoint), ``search`` (the query string
            corresponding to the search request), ``items_per_page`` (the number of
            items to return per page), ``skip`` (the number of items to skip) and
            ``page`` (which page to return).

          - *total_items_nb_key_path*: (optional) An XPath or JsonPath leading to the
            total number of results satisfying a request. This is used for providers
            which provides the total results metadata along with the result of the
            query and don't have an endpoint for querying the number of items
            satisfying a request, or for providers for which the count endpoint returns
            a json or xml document

          - *count_endpoint*: (optional) The endpoint for counting the number of items
            satisfying a request

        - **free_text_search_operations**: (optional) A tree structure of the form::

            <search-param>:     # e.g: $search
                union: # how to join the operations below (e.g: ' AND ' -->
                       # '(op1 AND op2) AND (op3 OR op4)')
                wrapper: # a pattern for how each operation will be wrapped
                         # (e.g: '({})' --> '(op1 AND op2)')
                operations:     # The operations to build
                  <opname>:     # e.g: AND
                    - <op1>    # e.g:
                               # 'sensingStartDate:[{startTimeFromAscendingNode}Z TO *]'
                    - <op2>    # e.g:
                        # 'sensingStopDate:[* TO {completionTimeFromAscendingNode}Z]'
                    ...
                  ...
            ...

          With the structure above, each operation will become a string of the form:
          '(<op1> <opname> <op2>)', then the operations will be joined together using
          the union string and finally if the number of operations is greater than 1,
          they will be wrapped as specified by the wrapper config key.

    The search plugins of this kind can detect when a metadata mapping is "query-able",
    and get the semantics of how to format the query string parameter that enables to
    make a query on the corresponding metadata. To make a metadata query-able, just
    configure it in the metadata mapping to be a list of 2 items, the first one being
    the specification of the query string search formatting. The later is a string
    following the specification of Python string formatting, with a special behaviour
    added to it. For example, an entry in the metadata mapping of this kind::

        completionTimeFromAscendingNode:
            - 'f=acquisition.endViewingDate:lte:{completionTimeFromAscendingNode#timestamp}'
            - '$.properties.acquisition.endViewingDate'

    means that the search url will have a query string parameter named *"f"* with a
    value of *"acquisition.endViewingDate:lte:1543922280.0"* if the search was done
    with the value of ``completionTimeFromAscendingNode`` being
    ``2018-12-04T12:18:00``. What happened is that
    ``{completionTimeFromAscendingNode#timestamp}`` was replaced with the timestamp
    of the value of ``completionTimeFromAscendingNode``. This example shows all there
    is to know about the semantics of the query string formatting introduced by this
    plugin: any eodag search parameter can be referenced in the query string
    with an additional optional conversion function that is separated from it by a
    ``#`` (see :func:`~eodag.utils.format_metadata` for further details on the
    available converters). Note that for the values in the
    ``free_text_search_operations`` configuration parameter follow the same rule.
    """

    COMPLEX_QS_REGEX = re.compile(r"^(.+=)?([^=]*)({.+})+([^=&]*)$")
    DEFAULT_ITEMS_PER_PAGE = 10
    extract_properties = {"xml": properties_from_xml, "json": properties_from_json}

    def __init__(self, provider, config):
        super(QueryStringSearch, self).__init__(provider, config)
        self.config.__dict__.setdefault("result_type", "json")
        self.config.__dict__.setdefault("results_entry", "features")
        self.config.__dict__.setdefault("pagination", {})
        self.config.__dict__.setdefault("free_text_search_operations", {})
        self.search_urls = []
        self.query_params = dict()
        self.query_string = ""

    def query(self, items_per_page=None, page=None, *args, **kwargs):
        """Perform a search on an OpenSearch-like interface"""
        product_type = kwargs.get("productType", None)
        provider_product_type = self.map_product_type(product_type, *args, **kwargs)
        keywords = {k: v for k, v in kwargs.items() if k != "auth" and v is not None}
        keywords["productType"] = provider_product_type
        qp, qs = self.build_query_string(product_type, *args, **keywords)
        # If we were not able to build query params but have search criteria, this means
        # the provider does not support the search criteria given. If so, stop searching
        # right away
        if not qp and keywords:
            return [], 0
        self.query_params = qp
        self.query_string = qs
        self.search_urls, total_items = self.collect_search_urls(
            page=page, items_per_page=items_per_page, *args, **kwargs
        )
        provider_results = self.do_search(
            items_per_page=items_per_page, *args, **kwargs
        )
        eo_products = self.normalize_results(
            provider_results, product_type, provider_product_type, *args, **kwargs
        )
        return eo_products, (total_items or len(eo_products))

    def build_query_string(self, *args, **kwargs):
        """Build The query string using the search parameters"""
        logger.debug("Building the query string that will be used for search")
        queryables = self.get_queryables(kwargs)
        query_params = {}
        # Get all the search parameters that are recognised as queryables by the
        # provider (they appear in the queryables dictionary)

        for eodag_search_key, provider_search_key in queryables.items():
            user_input = kwargs[eodag_search_key]

            if self.COMPLEX_QS_REGEX.match(provider_search_key):
                parts = provider_search_key.split("=")
                if len(parts) == 1:
                    query_params[eodag_search_key] = format_metadata(
                        provider_search_key
                    )
                else:
                    provider_search_key, provider_value = parts
                    query_params.setdefault(provider_search_key, []).append(
                        format_metadata(provider_value, *args, **kwargs)
                    )
            elif eodag_search_key == "custom":
                # extracts custom key:values from custom query parameter
                logger.debug('Converting custom query parameter "%s"' % user_input)
                custom_dict = parse_qs(user_input)
                # parsed dict as {k1:[v1], k2:[v2,v3]}
                for k in custom_dict.keys():
                    for custom_val in custom_dict[k]:
                        if k in query_params.keys():
                            query_params[k].append(custom_val)
                        else:
                            query_params.setdefault(k, []).append(custom_val)
            else:
                query_params[provider_search_key] = user_input

        # Now get all the literal search params (i.e params to be passed "as is"
        # in the search request)
        # ignore additional_params if it isn't a dictionary
        literal_search_params = getattr(self.config, "literal_search_params", {})
        if not isinstance(literal_search_params, dict):
            literal_search_params = {}

        # Now add formatted free text search parameters (this is for cases where a
        # complex query through a free text search parameter is available for the
        # provider and needed for the consumer)
        literal_search_params.update(self.format_free_text_search(**kwargs))
        for provider_search_key, provider_value in literal_search_params.items():
            if isinstance(provider_value, list):
                query_params.setdefault(provider_search_key, []).extend(provider_value)
            else:
                query_params.setdefault(provider_search_key, []).append(provider_value)

        # Build the final query string, in one go without quoting it
        # (some providers do not operate well with urlencoded and quoted query strings)
        return (
            query_params,
            urlencode(
                query_params, doseq=True, quote_via=lambda x, *_args, **_kwargs: x
            ),
        )

    def format_free_text_search(self, **kwargs):
        """Build the free text search parameter using the search parameters"""
        query_params = {}
        for param, operations_config in self.config.free_text_search_operations.items():
            union = operations_config["union"]
            wrapper = operations_config.get("wrapper", "{}")
            formatted_query = []
            for operator, operands in operations_config["operations"].items():
                # The Operator string is the operator wrapped with spaces
                operator = " {} ".format(operator)
                # Build the operation string by joining the formatted operands together
                # using the operation string
                operation_string = operator.join(
                    format_metadata(operand, **kwargs)
                    for operand in operands
                    if any(
                        kw in operand and val is not None for kw, val in kwargs.items()
                    )
                )
                # Finally wrap the operation string as specified by the wrapper and add
                # it to the list of queries (only if the operation string is not empty)
                if operation_string:
                    query = wrapper.format(operation_string)
                    formatted_query.append(query)
            # Join the formatted query using the "union" config parameter, and then
            # wrap it with the Python format string specified in the "wrapper" config
            # parameter
            final_query = union.join(formatted_query)
            if len(operations_config["operations"]) > 1 and len(formatted_query) > 1:
                final_query = wrapper.format(query_params[param])
            if final_query:
                query_params[param] = final_query
        return query_params

    def get_queryables(self, search_params):
        """Retrieve the metadata mappings that are query-able"""
        logger.debug("Retrieving queryable metadata from metadata_mapping")
        queryables = {}
        for eodag_search_key, user_input in search_params.items():
            if user_input is not None:
                md_mapping = self.config.metadata_mapping.get(eodag_search_key, None)
                if md_mapping is not None and isinstance(md_mapping, list):
                    search_param = get_search_param(md_mapping)
                    if search_param is not None:
                        queryables[eodag_search_key] = search_param
        return queryables

    def collect_search_urls(self, page=None, items_per_page=None, *args, **kwargs):
        """Build paginated urls"""
        urls = []
        total_results = 0
        for collection in self.get_collections(*args, **kwargs):
            search_endpoint = self.config.api_endpoint.rstrip("/").format(
                collection=collection
            )
            if page is not None and items_per_page is not None:
                count_endpoint = self.config.pagination.get(
                    "count_endpoint", ""
                ).format(collection=collection)
                if count_endpoint:
                    count_url = "{}?{}".format(count_endpoint, self.query_string)
                    _total_results = self.count_hits(
                        count_url, result_type=self.config.result_type
                    )
                else:
                    # First do one request querying only one element (lightweight
                    # request to schedule the pagination)
                    next_url_tpl = self.config.pagination["next_page_url_tpl"]
                    count_url = next_url_tpl.format(
                        url=search_endpoint,
                        search=self.query_string,
                        items_per_page=1,
                        page=1,
                        skip=0,
                    )
                    _total_results = self.count_hits(
                        count_url, result_type=self.config.result_type
                    )
                total_results += _total_results or 0
                next_url = self.config.pagination["next_page_url_tpl"].format(
                    url=search_endpoint,
                    search=self.query_string,
                    items_per_page=items_per_page,
                    page=page,
                    skip=(page - 1) * items_per_page,
                )
            else:
                next_url = "{}?{}".format(search_endpoint, self.query_string)
            urls.append(next_url)
        return urls, total_results

    def do_search(self, items_per_page=None, *args, **kwargs):
        """Perform the actual search request.

        If there is a specified number of items per page, return the results as soon
        as this number is reached

        :param int items_per_page: (Optional) The number of items to return for one page
        """
        results = []
        for search_url in self.search_urls:
            try:
                response = self._request(
                    search_url,
                    info_message="Sending search request: {}".format(search_url),
                    exception_message="Skipping error while searching for {} {} "
                    "instance:".format(self.provider, self.__class__.__name__),
                )
            except RequestError:
                raise StopIteration
            else:
                if self.config.result_type == "xml":
                    root_node = etree.fromstring(response.content)
                    namespaces = {k or "ns": v for k, v in root_node.nsmap.items()}
                    results.extend(
                        [
                            etree.tostring(entry)
                            for entry in root_node.xpath(
                                self.config.results_entry, namespaces=namespaces
                            )
                        ]
                    )
                else:
                    results.extend(response.json().get(self.config.results_entry, []))
            if items_per_page is not None and len(results) == items_per_page:
                return results
        return results

    def normalize_results(self, results, *args, **kwargs):
        """Build EOProducts from provider results"""
        logger.debug("Adapting plugin results to eodag product representation")
        return [
            EOProduct(
                self.provider,
                QueryStringSearch.extract_properties[self.config.result_type](
                    result, self.config.metadata_mapping
                ),
                *args,
                **kwargs
            )
            for result in results
        ]

    def count_hits(self, count_url, result_type="json"):
        """Count the number of results satisfying some criteria"""
        # Handle a very annoying special case :'(
        url = count_url.replace("$format=json&", "")
        response = self._request(
            url,
            info_message="Sending count request: {}".format(url),
            exception_message="Skipping error while counting results for {} {} "
            "instance:".format(self.provider, self.__class__.__name__),
        )
        if result_type == "xml":
            root_node = etree.fromstring(response.content)
            total_nb_results = root_node.xpath(
                self.config.pagination["total_items_nb_key_path"],
                namespaces={k or "ns": v for k, v in root_node.nsmap.items()},
            )[0]
            total_results = int(total_nb_results)
        else:
            count_results = response.json()
            if isinstance(count_results, dict):
                path_parsed = jsonpath.parse(
                    self.config.pagination["total_items_nb_key_path"]
                )
                total_results = path_parsed.find(count_results)[0].value
            else:  # interpret the result as a raw int
                total_results = int(count_results)
        return total_results

    def get_collections(self, *args, **kwargs):
        """Get the collection to which the product belongs"""
        # See https://earth.esa.int/web/sentinel/missions/sentinel-2/news/-
        # /asset_publisher/Ac0d/content/change-of
        # -format-for-new-sentinel-2-level-1c-products-starting-on-6-december
        product_type = kwargs.get("productType")
        if product_type is None:
            collections = set()
            collection = getattr(self.config, "collection", None)
            if collection is None:
                try:
                    for product_type, product_config in self.config.products.items():
                        collections.add(product_config["collection"])
                except KeyError:
                    collections.add("")
            else:
                collections.add(collection)
            return tuple(collections)
        if self.provider == "peps":
            if product_type == "S2_MSI_L1C":
                date = kwargs.get("startTimeFromAscendingNode")
                # If there is no criteria on date, we want to query all the collections
                # known for providing L1C products
                if date is None:
                    collections = ("S2", "S2ST")
                else:
                    match = re.match(
                        r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})", date
                    ).groupdict()
                    year, month, day = (
                        int(match["year"]),
                        int(match["month"]),
                        int(match["day"]),
                    )
                    if year > 2016 or (year == 2016 and month == 12 and day > 5):
                        collections = ("S2ST",)
                    else:
                        collections = ("S2", "S2ST")
            else:
                collections = (
                    self.config.products[product_type].get("collection", ""),
                )
        else:
            collection = getattr(self.config, "collection", None)
            if collection is None:
                collection = self.config.products[product_type].get("collection", "")
            collections = (collection,)
        return collections

    def map_product_type(self, product_type, *args, **kwargs):
        """Map the eodag product type to the provider product type"""
        if product_type is None:
            return
        logger.debug("Mapping eodag product type to provider product type")
        return self.config.products[product_type]["product_type"]

    def _request(self, url, info_message=None, exception_message=None):
        try:
            if info_message:
                logger.info(info_message)
            response = requests.get(url)
            response.raise_for_status()
        except requests.HTTPError as err:
            if exception_message:
                logger.exception(exception_message)
            else:
                logger.exception(
                    "Skipping error while requesting: %s (provider:%s, plugin:%s):",
                    url,
                    self.provider,
                    self.__class__.__name__,
                )
            raise RequestError(str(err))
        return response


class AwsSearch(QueryStringSearch):
    """A specialisation of RestoSearch that modifies the way the EOProducts are built
    from the search results"""

    def normalize_results(self, results, *args, **kwargs):
        """Transform metadata from provider representation to eodag representation"""
        normalized = []
        logger.debug("Adapting plugin results to eodag product representation")
        for result in results:
            ref = result["properties"]["title"].split("_")[5]
            year = result["properties"]["completionDate"][0:4]
            month = str(int(result["properties"]["completionDate"][5:7]))
            day = str(int(result["properties"]["completionDate"][8:10]))

            properties = QueryStringSearch.extract_properties[self.config.result_type](
                result, self.config.metadata_mapping
            )

            properties["downloadLink"] = (
                "s3://tiles/{ref[1]}{ref[2]}/{ref[3]}/{ref[4]}{ref[5]}/{year}/"
                "{month}/{day}/0/"
            ).format(**locals())
            normalized.append(EOProduct(self.provider, properties, *args, **kwargs))
        return normalized


class ODataV4Search(QueryStringSearch):
    """A specialisation of a QueryStringSearch that does a two step search to retrieve
    all products metadata"""

    def do_search(self, *args, **kwargs):
        """Do a two step search, as the metadata are not given into the search result"""
        # TODO: This plugin is still very specific to the ONDA provider.
        #       Be careful to generalize it if needed when the chance to do so arrives
        final_result = []
        # Query the products entity set for basic metadata about the product
        for entity in super(ODataV4Search, self).do_search(*args, **kwargs):
            if entity["downloadable"]:
                entity_metadata = {
                    "quicklook": entity["quicklook"],
                    "id": entity["id"],
                    "footprint": entity["footprint"],
                }
                metadata_url = self.get_metadata_search_url(entity)
                try:
                    response = requests.get(metadata_url)
                    response.raise_for_status()
                except requests.HTTPError:
                    logger.exception(
                        "Skipping error while searching for %s %s instance:",
                        self.provider,
                        self.__class__.__name__,
                    )
                else:
                    entity_metadata.update(
                        {item["id"]: item["value"] for item in response.json()["value"]}
                    )
                    final_result.append(entity_metadata)
        return final_result

    def get_metadata_search_url(self, entity):
        """Build the metadata link for the given entity"""
        return "{}({})/Metadata".format(
            self.config.api_endpoint.rstrip("/"), entity["id"]
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import jsonpath_rw as jsonpath

from eodag.api.product.metadata_mapping import DEFAULT_METADATA_MAPPING, NOT_MAPPED, get_metadata_path
from eodag.plugins.base import PluginTopic


class Search(PluginTopic):

    def __init__(self, provider, config):
        super(Search, self).__init__(provider, config)
        # Prepare the metadata mapping
        # Do a shallow copy, the structure is flat enough for this to be sufficient
        metas = DEFAULT_METADATA_MAPPING.copy()
        # Update the defaults with the mapping value. This will add any new key
        # added by the provider mapping that is not in the default metadata
        metas.update(self.config.metadata_mapping)
        for metadata in metas:
            if metadata not in self.config.metadata_mapping:
                self.config.metadata_mapping[metadata] = (None, NOT_MAPPED)
            else:
                conversion, path = get_metadata_path(self.config.metadata_mapping[metadata])
                try:
                    # If the metadata is queryable (i.e a list of 2 elements), replace the value of the last item
                    if len(self.config.metadata_mapping[metadata]) == 2:
                        self.config.metadata_mapping[metadata][1] = (conversion, jsonpath.parse(path))
                    else:
                        self.config.metadata_mapping[metadata] = (conversion, jsonpath.parse(path))
                except Exception:  # jsonpath_rw does not provide a proper exception
                    # Assume the mapping is to be passed as is.
                    # Ignore any transformation specified. If a value is to be passed as is, we don't want to transform
                    # it further
                    _, text = get_metadata_path(self.config.metadata_mapping[metadata])
                    if len(self.config.metadata_mapping[metadata]) == 2:
                        self.config.metadata_mapping[metadata][1] = (None, text)
                    else:
                        self.config.metadata_mapping[metadata] = (None, text)

    def query(self, *args, **kwargs):
        """Implementation of how the products must be searched goes here.

        This method must return a list of EOProduct instances (see eodag.api.product module) which will
        be processed by a Download plugin.
        """
        raise NotImplementedError('A Search plugin must implement a method named query')



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from eodag.plugins.authentication.base import Authentication
from eodag.utils.exceptions import MisconfiguredError


class GenericAuth(Authentication):

    def authenticate(self):
        method = getattr(self.config, 'method', None)
        try:
            if not method:
                method = 'basic'
            if method == 'basic':
                return HTTPBasicAuth(
                    self.config.credentials['username'],
                    self.config.credentials['password']
                )
            if method == 'digest':
                return HTTPDigestAuth(
                    self.config.credentials['username'],
                    self.config.credentials['password']
                )
        except AttributeError as err:
            if 'credentials' in err.args:
                raise MisconfiguredError('Missing Credentials for provider: %s', self.provider)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import six


class DatasetDriver(six.with_metaclass(type)):

    def get_data_address(self, eo_product, band):
        """Retrieve the address of the dataset represented by `eo_product`.

        :param eo_product: The product whom underlying dataset address is to be retrieved
        :type eo_product: :class:`~eodag.api.product.EOProduct`
        :param band: The band to retrieve (e.g: 'B01')
        :type band: str or unicode
        :returns: An address for the dataset
        :rtype: str or unicode
        :raises: :class:`~eodag.utils.exceptions.AddressNotFound`
        :raises: :class:`~eodag.utils.exceptions.UnsupportedDatasetAddressScheme`
        """
        raise NotImplementedError


class NoDriver(DatasetDriver):
    """A default driver that does not implement any of the methods it should implement, used for all product types for
    which the :meth:`~eodag.api.product._product.EOProduct.get_data` method is not yet implemented in eodag. Expect a
    :exc:`NotImplementedError` when trying to get the data in that case.
    """



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals


try:  # PY2
    from UserList import UserList
except ImportError:  # PY3
    from collections import UserList

from eodag.api.product import EOProduct


class SearchResult(UserList):
    """An object representing a collection of :class:`~eodag.api.product.EOProduct` resulting from a search.

    :param products: A list of products resulting from a search
    :type products: list(:class:`~eodag.api.product.EOProduct`)
    """

    def __init__(self, products):
        super(SearchResult, self).__init__(products)

    def crunch(self, cruncher, **search_params):
        """Do some crunching with the underlying EO products.

        :param cruncher: The plugin instance to use to work on the products
        :type cruncher: subclass of :class:`~eodag.plugins.crunch.base.Crunch`
        :param dict search_params: The criteria that have been used to produce this result
        :returns: The result of the application of the crunching method to the EO products
        :rtype: :class:`~eodag.api.search_result.SearchResult`
        """
        crunched_results = cruncher.proceed(self, **search_params)
        return SearchResult(crunched_results)

    @staticmethod
    def from_geojson(feature_collection):
        """Builds an :class:`~eodag.api.search_result.SearchResult` object from its representation as geojson

        :param feature_collection: A collection representing a search result.
        :type feature_collection: dict
        :returns: An eodag representation of a search result
        :rtype: :class:`~eodag.api.search_result.SearchResult`
        """
        return SearchResult(EOProduct.from_geojson(feature) for feature in feature_collection['features'])

    def as_geojson_object(self):
        return {'type': 'FeatureCollection', 'features': [product.as_dict() for product in self]}

    @property
    def __geo_interface__(self):
        """Implements the geo-interface protocol.

        See https://gist.github.com/sgillies/2217756
        """
        return self.as_geojson_object()



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import logging
import time

import grpc
from concurrent.futures import ThreadPoolExecutor

from eodag import EODataAccessGateway
from eodag.rpc.protocol import eodag_pb2, eodag_pb2_grpc

logger = logging.getLogger("eodag.rpc.server")


class EODAGRPCServer(object):
    """Eodag grpc server definition"""

    ONE_DAY_IN_SECONDS = 60 * 60 * 24

    def __init__(self, host, port, user_conf):
        logger.info("Initializing eodag rpc server")
        eodag = EODataAccessGateway(user_conf_file_path=user_conf)
        self.eo_product_type_iface = EOProductTypeService(eodag)
        self.eo_product_iface = EOProductService(eodag)
        self.address = "{}:{}".format(host, port)

    def serve(self):
        """Expose eodag functionality"""
        server = grpc.server(ThreadPoolExecutor(max_workers=10))
        eodag_pb2_grpc.add_EOProductServicer_to_server(self.eo_product_iface, server)
        eodag_pb2_grpc.add_EOProductTypeServicer_to_server(
            self.eo_product_type_iface, server
        )
        server.add_insecure_port(self.address)
        logger.info("Starting eodag rpc server")
        server.start()
        logger.info(
            "eodag rpc server is now listening for requests at %s", self.address
        )
        try:
            while True:
                time.sleep(self.ONE_DAY_IN_SECONDS)
        except KeyboardInterrupt:
            server.stop(0)


class EOProductTypeService(eodag_pb2_grpc.EOProductTypeServicer):
    """EODAG grpc service for product types"""

    def __init__(self, eodag):
        super(EOProductTypeService, self).__init__()
        self.eodag = eodag

    def ListProductTypes(self, request, context):
        """List the product types known by the system"""
        logger.info("New request for available product types: %s", request or "<Empty>")
        logger.debug("Context: %s", context)
        for product_type in self.eodag.list_product_types():
            schema = eodag_pb2.EOProductTypeSchema()
            schema.id = product_type["ID"]
            schema.description = product_type["desc"]
            for metadata in product_type["meta"]:
                meta = schema.meta.add()
                meta.key = metadata["key"]
                meta.value_type = metadata["type"]
            yield schema
        logger.info("Request successfully executed")

    def SearchProductType(self, request, context):
        """Show the information on a known product type"""
        # TODO: implement the interaction with whoosh here


class EOProductService(eodag_pb2_grpc.EOProductServicer):
    """EODAG grpc product service"""

    def __init__(self, eodag):
        super(EOProductService, self).__init__()
        self.eodag = eodag

    def SearchProduct(self, request, context):
        """Search a product"""
        pass

    def DownloadProduct(self, request, context):
        """Download a product"""
        pass



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

__title__ = "eodag"
__description__ = "Earth Observation Data Access Gateway"
__version__ = "1.3.3"
__author__ = "CS Systemes d'Information (CSSI)"
__author_email__ = "admin@geostorm.eu"
__url__ = "https://bitbucket.org/geostorm/eodag"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2018, CS Systemes d'Information, http://www.c-s.fr"



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eo_product.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='eo_product.proto',
  package='eodag',
  syntax='proto3',
  serialized_pb=_b('\n\x10\x65o_product.proto\x12\x05\x65odag\">\n\x07NDArray\x12\x11\n\x05\x61rray\x18\x01 \x03(\x01\x42\x02\x10\x01\x12\x11\n\x05shape\x18\x02 \x03(\x07\x42\x02\x10\x01\x12\r\n\x05\x64type\x18\x03 \x01(\t\"\x89\x01\n\x13\x45OProductSubdataset\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08producer\x18\x02 \x01(\t\x12\x14\n\x0cproduct_type\x18\x03 \x01(\t\x12\x10\n\x08platform\x18\x04 \x01(\t\x12\x0e\n\x06sensor\x18\x05 \x01(\t\x12\x1c\n\x04\x64\x61ta\x18\x06 \x01(\x0b\x32\x0e.eodag.NDArrayb\x06proto3')
)




_NDARRAY = _descriptor.Descriptor(
  name='NDArray',
  full_name='eodag.NDArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='array', full_name='eodag.NDArray.array', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001')), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='eodag.NDArray.shape', index=1,
      number=2, type=7, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001')), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='eodag.NDArray.dtype', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=27,
  serialized_end=89,
)


_EOPRODUCTSUBDATASET = _descriptor.Descriptor(
  name='EOProductSubdataset',
  full_name='eodag.EOProductSubdataset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='eodag.EOProductSubdataset.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='producer', full_name='eodag.EOProductSubdataset.producer', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='product_type', full_name='eodag.EOProductSubdataset.product_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='platform', full_name='eodag.EOProductSubdataset.platform', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sensor', full_name='eodag.EOProductSubdataset.sensor', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='eodag.EOProductSubdataset.data', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=92,
  serialized_end=229,
)

_EOPRODUCTSUBDATASET.fields_by_name['data'].message_type = _NDARRAY
DESCRIPTOR.message_types_by_name['NDArray'] = _NDARRAY
DESCRIPTOR.message_types_by_name['EOProductSubdataset'] = _EOPRODUCTSUBDATASET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NDArray = _reflection.GeneratedProtocolMessageType('NDArray', (_message.Message,), dict(
  DESCRIPTOR = _NDARRAY,
  __module__ = 'eo_product_pb2'
  # @@protoc_insertion_point(class_scope:eodag.NDArray)
  ))
_sym_db.RegisterMessage(NDArray)

EOProductSubdataset = _reflection.GeneratedProtocolMessageType('EOProductSubdataset', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTSUBDATASET,
  __module__ = 'eo_product_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductSubdataset)
  ))
_sym_db.RegisterMessage(EOProductSubdataset)


_NDARRAY.fields_by_name['array'].has_options = True
_NDARRAY.fields_by_name['array']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_NDARRAY.fields_by_name['shape'].has_options = True
_NDARRAY.fields_by_name['shape']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import six

from eodag.utils.exceptions import PluginNotFoundError


class EODAGPluginMount(type):
    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, 'plugins'):
            # This branch only executes when processing the mount point itself.
            # So, since this is a new plugin type, not an implementation, this
            # class shouldn't be registered as a plugin. Instead, it sets up a
            # list where plugins can be registered later.
            cls.plugins = []
        else:
            # This must be a plugin implementation, which should be registered.
            # Simply appending it to the list is all that's needed to keep
            # track of it later.
            cls.plugins.append(cls)

    def get_plugins(cls, *args, **kwargs):
        return [plugin(*args, **kwargs) for plugin in cls.plugins]

    def get_plugin_by_class_name(cls, name):
        for plugin in cls.plugins:
            if name == plugin.__name__:
                return plugin
        raise PluginNotFoundError("'{}' not found for {} class of plugins".format(name, cls))


class PluginTopic(six.with_metaclass(EODAGPluginMount)):
    """Base of all plugin topics in eodag"""

    def __init__(self, provider, config):
        self.config = config
        self.provider = provider

    def __repr__(self):
        return '{}(provider={}, priority={}, topic={})'.format(
            self.__class__.__name__, self.provider, self.config.priority, self.__class__.mro()[-3].__name__
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ValidationError(Exception):
    """Error validating data"""

    def __init__(self, message):
        self.message = message


class PluginNotFoundError(Exception):
    """Error when looking for a plugin class that was not defined"""


class PluginImplementationError(Exception):
    """Error when a plugin does not behave as expected"""


class MisconfiguredError(Exception):
    """An error indicating a Search Plugin that is not well configured"""


class AddressNotFound(Exception):
    """An error indicating the address of a subdataset was not found"""


class UnsupportedProvider(Exception):
    """An error indicating that eodag does not support a provider"""


class UnsupportedProductType(Exception):
    """An error indicating that eodag does not support a product type"""

    def __init__(self, product_type):
        self.product_type = product_type


class UnsupportedDatasetAddressScheme(Exception):
    """An error indicating that eodag does not yet support an address scheme for
    accessing raster subdatasets"""


class AuthenticationError(Exception):
    """An error indicating that an authentication plugin did not succeeded
    authenticating a user"""


class DownloadError(Exception):
    """An error indicating something wrong with the download process"""


class RequestError(Exception):
    """An error indicating that a HTTP request has failed. Usually eodag functions
    and methods should catch and skip this"""


class NoMatchingProductType(Exception):
    """An error indicating that eodag was unable to derive a product type from a set
    of search parameters"""



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import os
import re
import warnings

import boto3
import rasterio

from eodag.api.product.drivers.base import DatasetDriver
from eodag.utils.exceptions import AddressNotFound, UnsupportedDatasetAddressScheme


class Sentinel2L1C(DatasetDriver):
    BAND_FILE_PATTERN_TPL = r'^.+_{band}\.jp2$'
    SPATIAL_RES_PER_BANDS = {
        '10m': ('B02', 'B03', 'B04', 'B08'),
        '20m': ('B05', 'B06', 'B07', 'B11', 'B12', 'B8A'),
        '60m': ('B01', 'B09', 'B10'),
        'TCI': ('TCI',),
    }

    def get_data_address(self, eo_product, band):
        """Compute the address of a subdataset for a Sentinel2 L1C product.

        The algorithm is as follows for a product on the local filesystem:
            - First compute the top level metadata file path by appending `MTD_MSIL1C.xml` to the
              ``eo_product.location``. Then open it as a `rasterio` dataset
            - Then mimics the shell command ``gdalinfo -sd n /path/metadata.xml`` to get the final address:
                - iterate through the subdataset addresses ('<DRIVER>:<path>/<mtd>.xml:<spatial-resolution>:<crs>')
                  detected by the rasterio dataset
                - open only the address for which the extracted spatial resolution maps to a tuple of bands including
                  the band of interest
            - Finally, filter the list of files of the previously opened rasterio dataset, to return the
              filesystem-like address that matches the band file pattern r'^.+_B01\\.jp2$' if band = 'B01' for
              example.

        See :func:`~eodag.api.product.drivers.base.DatasetDriver.get_data_address` to get help on the formal
        parameters.
        """
        product_location_scheme = eo_product.location.split('://')[0]
        if product_location_scheme == 'file':
            top_level_mtd = os.path.join(re.sub(r'file://', '', eo_product.location), 'MTD_MSIL1C.xml')
            # Ignore the NotGeoreferencedWarning thrown by rasterio
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=UserWarning, message='Dataset has no geotransform set')
                with rasterio.open(top_level_mtd) as dataset:
                    for address in dataset.subdatasets:
                        spatial_res = address.split(':')[-2]
                        if band in self.SPATIAL_RES_PER_BANDS[spatial_res]:
                            with rasterio.open(address) as subdataset:
                                band_file_pattern = re.compile(self.BAND_FILE_PATTERN_TPL.format(band=band))
                                for filename in filter(lambda f: band_file_pattern.match(f), subdataset.files):
                                    return filename
                raise AddressNotFound
        if product_location_scheme == 's3':
            access_key, access_secret = eo_product.downloader_auth.authenticate()
            s3 = boto3.resource('s3', aws_access_key_id=access_key, aws_secret_access_key=access_secret)
            bucket = s3.Bucket('sentinel-s2-l1c')
            for summary in bucket.objects.filter(Prefix=eo_product.location.split('s3://')[-1]):
                if '{}.jp2'.format(band) in summary.key:
                    return 's3://sentinel-s2-l1c/{}'.format(summary.key)
            raise AddressNotFound
        raise UnsupportedDatasetAddressScheme('eo product {} is accessible through a location scheme that is not yet '
                                              'supported by eodag: {}'.format(eo_product, product_location_scheme))



#!/usr/bin/env python

# -*- coding: utf-8 -*-
# Copyright 2017-2018 CS Systemes d'Information (CS SI)
# All rights reserved

from __future__ import absolute_import, unicode_literals

import os
import sys
from functools import wraps

import flask
import requests
from flask import jsonify, make_response, render_template, request

from eodag.api.core import DEFAULT_ITEMS_PER_PAGE
from eodag.rest.utils import (
    get_home_page_content,
    get_product_types,
    search_product_by_id,
    search_products,
)
from eodag.utils import makedirs
from eodag.utils.exceptions import (
    UnsupportedProductType,
    UnsupportedProvider,
    ValidationError,
)

app = flask.Flask(__name__)


def cross_origin(request_handler):
    """Wraps a view to relax the need for csrf token"""

    @wraps(request_handler)
    def wrapper(*args, **kwargs):
        resp = make_response(request_handler(*args, **kwargs))
        resp.headers["Access-Control-Allow-Origin"] = "*"
        return resp

    return wrapper


@app.route("/<product_type>/", methods=["GET"])
@cross_origin
def search(product_type):
    """Search for a product by product type on eodag"""
    try:
        response = search_products(product_type, request.args)
    except ValidationError as e:
        return jsonify({"error": e.message}), 400
    except RuntimeError as e:
        return jsonify({"error": e}), 400
    except UnsupportedProductType as e:
        return jsonify({"error": "Not Found: {}".format(e.product_type)}), 404

    return jsonify(response), 200


@app.route("/search/<uid>/", methods=["GET"])
@cross_origin
def search_by_id(uid):
    """Retrieve the quicklook of a eo product identified by its id"""
    provider = request.args.get("provider")
    try:
        search_result = search_product_by_id(uid, provider=provider)
    except ValidationError as e:
        return jsonify({"error": e.message}), 400
    except RuntimeError as e:
        return jsonify({"error": e}), 500
    except UnsupportedProvider:
        return jsonify({"error": "Unknown provider: %s" % (provider,)}), 400

    if len(search_result) == 0:
        return jsonify({"error": "Not found"}), 404
    return jsonify(search_result[0].as_dict()), 200


@app.route("/quicklook/<uid>/", methods=["GET"])
@cross_origin
def get_quicklook(uid=None):
    """Retrieve the quicklook of a eo product identified by its id"""
    if uid is None:
        return jsonify({"error": "You must provide a EO product uid"}), 400
    provider = request.args.get("provider")
    try:
        search_result = search_product_by_id(uid, provider=provider)
    except ValidationError as e:
        return jsonify({"error": e.message}), 400
    except RuntimeError as e:
        return jsonify({"error": e}), 500
    except UnsupportedProvider:
        return jsonify({"error": "Unknown provider: %s" % (provider,)}), 400

    if len(search_result) == 0:
        return jsonify({"error": "EO product not found"}), 400

    eo_product = search_result[0]
    quicklook = eo_product.properties.get("quicklook", None)
    if quicklook is None:
        return jsonify({"error": "Not Found"}), 404
    quicklook_alt_text = "{} quicklook".format(uid)
    if quicklook.startswith("http") or quicklook.startswith("https"):
        # First see if the url of the quicklook is accessible as is
        resp = requests.get(quicklook)
        try:
            resp.raise_for_status()
            quicklook_src = quicklook
        except requests.HTTPError:
            # Flask is known to automatically serve files present in a folder named
            # "static" in the root folder of the application.
            # If the url is not accessible as is, try to get it from the provider using
            # its authentication mechanism
            quicklooks_dir = os.path.join(
                os.path.dirname(__file__), "static", "quicklooks"
            )
            # First create the static folder and quicklooks dir inside it if necessary
            makedirs(quicklooks_dir)
            # Then download the quicklook
            response = eo_product.get_quicklook(filename=uid, base_dir=quicklooks_dir)
            # get_quicklook should always return an absolute path, which starts with "/"
            # If it fails to do so, we consider an error occured while getting the
            # quicklook
            if not response.startswith("/"):
                return jsonify({"error": response}), 500
            quicklook_src = "/static/quicklooks/{}".format(uid)
    # If the quicklook is not an HTTP URL, we guess it is a base64 stream. In that case
    # we directly include it in the <img> tag and return it as is
    else:
        quicklook_src = "data:image/png;base64, {}".format(quicklook)
    return '<img src="{}" alt="{}" />'.format(quicklook_src, quicklook_alt_text), 200


@app.route("/", methods=["GET"])
@cross_origin
def home():
    """Render home page"""
    return render_template(
        "index.html",
        content=get_home_page_content(request.base_url, DEFAULT_ITEMS_PER_PAGE),
    )


@app.route("/product-types/", methods=["GET"])
@app.route("/product-types/<provider>", methods=["GET"])
@cross_origin
def list_product_types(provider=None):
    """List eodag' supported product types"""
    try:
        product_types = get_product_types(provider, request.args)
    except UnsupportedProvider:
        return jsonify({"error": "Unknown provider: %s" % (provider,)}), 400
    except Exception:
        return jsonify({"error": "Unknown server error"}), 500
    return jsonify(product_types), 200


def main():
    """Launch the server"""
    import argparse

    parser = argparse.ArgumentParser(
        description="""Script for starting EODAG server""", epilog=""""""
    )
    parser.add_argument(
        "-d", "--daemon", action="store_true", help="run in daemon mode"
    )
    parser.add_argument(
        "-a",
        "--all-addresses",
        action="store_true",
        help="run flask using IPv4 0.0.0.0 (all network interfaces), "
        "otherwise bind to 127.0.0.1 (localhost). "
        "This maybe necessary in systems that only run Flask",
    )
    args = parser.parse_args()

    if args.all_addresses:
        bind_host = "0.0.0.0"
    else:
        bind_host = "127.0.0.1"

    if args.daemon:
        pid = None
        try:
            pid = os.fork()
        except OSError as e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))

        if pid == 0:
            os.setsid()
            app.run(threaded=True, host=bind_host)
        else:
            sys.exit(0)
    else:
        # For development
        app.run(debug=True, use_reloader=True)


if __name__ == "__main__":
    main()
